import pytest
from typer.testing import CliRunner

from lib.test_lib import assert_lines_match, path_in_test_data, run_and_report

runner = CliRunner()


PALES_TEMPLATE = ["pales", "export", "rdcs"]
SEQUENCE_STREAM = open(
    path_in_test_data(__file__, "pales_test_1.nef", local=True)
).read()
SEQUENCE_STREAM_DISORDERED = open(
    path_in_test_data(__file__, "pales_test_2.nef", local=True)
).read()


@pytest.fixture
def using_pales():
    # register the module under test
    import transcoders.pales  # noqa: F401


# noinspection PyUnusedLocal
def test_rdcs(typer_app, using_pales, monkeypatch, clear_cache):

    monkeypatch.setattr("sys.stdin.isatty", lambda: False)

    result = runner.invoke(typer_app, [*PALES_TEMPLATE], input=SEQUENCE_STREAM)

    if result.exit_code != 0:
        print("INFO: stdout from failed read:\n", result.stdout)

    assert result.exit_code == 0

    EXPECTED = """\
        REMARK NEF CHAIN A
        REMARK NEF START RESIDUE 1

        DATA SEQUENCE AWG

        VARS    RESID_I  RESNAME_I  ATOMNAME_I  RESID_J  RESNAME_J  ATOMNAME_J  D      DD     W
        FORMAT  %5d      %6s        %6s         %5d      %6s        %6s         %9.3f  %9.3f  %.2f
                2        TRP        HN          21       TRP        N          -5.2    0.33   1.0
                3        GLY        HN          22       GLY        N           3.1    0.4    1.0

    """

    assert_lines_match(EXPECTED, result.stdout)


# noinspection PyUnusedLocal
def test_rdcs_disordered(typer_app, using_pales, monkeypatch, clear_cache):
    monkeypatch.setattr("sys.stdin.isatty", lambda: False)

    result = run_and_report(
        typer_app, [*PALES_TEMPLATE], input=SEQUENCE_STREAM_DISORDERED
    )

    assert result.exit_code == 0

    EXPECTED = """\
        REMARK NEF CHAIN A
        REMARK NEF START RESIDUE 1

        DATA SEQUENCE AWG

        VARS    RESID_I  RESNAME_I  ATOMNAME_I  RESID_J  RESNAME_J  ATOMNAME_J  D      DD     W
        FORMAT  %5d      %6s        %6s         %5d      %6s        %6s         %9.3f  %9.3f  %.2f
                2        TRP        HN          21       TRP        N          -5.2    0.33   1.0
                3        GLY        HN          22       GLY        N           3.1    0.4    1.0

    """

    assert_lines_match(EXPECTED, result.stdout)



import traceback
from textwrap import dedent

import pytest
from icecream import ic

from typer.testing import CliRunner

from lib.test_lib import assert_lines_match, isolate_frame, path_in_test_data, run_and_report

HEADER = open(path_in_test_data(__file__,'test_header_entry.txt', local=False)).read()

PEAKS_NMRPIPE = 'nef_nmr_spectrum_nmrpipe'
METADATA_NMRPIPE = 'nef_nmr_meta_data'
NMRPIPE_IMPORT_PEAKS = ['nmrpipe', 'import', 'peaks']

runner = CliRunner()

@pytest.fixture
def using_nmrpipe():
    # register the module under test
    import transcoders.nmrpipe


# noinspection PyUnusedLocal
def test_peaks(typer_app, using_nmrpipe, monkeypatch):
    EXPECTED = open(path_in_test_data(__file__, 'gb3_assigned_trunc_expected.tab')).read()
    monkeypatch.setattr('sys.stdin.isatty', lambda: False)

    path = path_in_test_data(__file__, 'gb3_assigned_trunc.tab')
    result = run_and_report(typer_app, [*NMRPIPE_IMPORT_PEAKS, path], input=HEADER)

    assert result.exit_code == 0
    peaks_result = isolate_frame(result.stdout, '%s' % PEAKS_NMRPIPE)

    assert_lines_match(EXPECTED, peaks_result)


if __name__ == '__main__':
    pytest.main(['--disable-warnings', '-vv', f'{__file__}'])



from textwrap import dedent

import pytest
from typer.testing import CliRunner

from lib.structures import AtomLabel, SequenceResidue, ShiftData, ShiftList
from lib.test_lib import (
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    run_and_report,
)
from transcoders.nmrview.nmrview_lib import parse_shifts

runner = CliRunner()

SHIFTS_NMRPIPE = "nef_chemical_shift_list_nmrview"
METADATA_NMRVIEW = "nef_nmr_meta_data"
NMRVIEW_IMPORT_SHIFTS = ["nmrview", "import", "shifts"]


@pytest.fixture
def using_nmrview():
    # register the module under test
    import transcoders.nmrview  # noqa: F401


def test_lib_parse_shifts():

    EXPECTED = ShiftList(
        shifts=[
            ShiftData(
                atom=AtomLabel(
                    SequenceResidue(
                        chain_code="A", sequence_code=1, residue_name="ASP"
                    ),
                    atom_name="CA",
                ),
                shift=52.00,
                error=None,
            ),
            ShiftData(
                atom=AtomLabel(
                    SequenceResidue(
                        chain_code="A", sequence_code=1, residue_name="ASP"
                    ),
                    atom_name="HA",
                ),
                shift=4.220,
                error=None,
            ),
            ShiftData(
                atom=AtomLabel(
                    SequenceResidue(
                        chain_code="A", sequence_code=2, residue_name="VAL"
                    ),
                    atom_name="CG2",
                ),
                shift=19.300,
                error=None,
            ),
            ShiftData(
                atom=AtomLabel(
                    SequenceResidue(
                        chain_code="A", sequence_code=2, residue_name="VAL"
                    ),
                    atom_name="HG21",
                ),
                shift=0.814,
                error=None,
            ),
            ShiftData(
                atom=AtomLabel(
                    SequenceResidue(
                        chain_code="A", sequence_code=2, residue_name="VAL"
                    ),
                    atom_name="HG22",
                ),
                shift=0.814,
                error=None,
            ),
            ShiftData(
                atom=AtomLabel(
                    SequenceResidue(
                        chain_code="A", sequence_code=3, residue_name="GLN"
                    ),
                    atom_name="N",
                ),
                shift=125.058,
                error=None,
            ),
        ]
    )

    test_data = """\
          1.CA      52.000 1
          1.HA       4.220 1
          2.CG2     19.300 1
          2.HG21     0.814 1
          2.HG22     0.814 1
          3.N      125.058 1
    """

    chain_seqid_to_type = {("A", 1): "ASP", ("A", 2): "VAL", ("A", 3): "GLN"}

    test_data = dedent(test_data)

    shifts = parse_shifts(test_data.split("\n"), chain_seqid_to_type)

    assert shifts == EXPECTED


#
EXPECTED_PPM_OUT_SHORT = """\
save_nef_chemical_shift_list_nmrview
   _nef_chemical_shift_list.sf_category   nef_chemical_shift_list
   _nef_chemical_shift_list.sf_framecode  nef_chemical_shift_list_nmrview

   loop_
      _nef_chemical_shift_list.chain_code
      _nef_chemical_shift_list.sequence_code
      _nef_chemical_shift_list.residue_name
      _nef_chemical_shift_list.atom_name
      _nef_chemical_shift_list.value
      _nef_chemical_shift_list.value_uncertainty
      _nef_chemical_shift_list.element
      _nef_chemical_shift_list.isotope_number

     A   1   ASP   CA     52.0      .   .   .
     A   1   ASP   HA     4.22      .   .   .
     A   2   VAL   CG2    19.3      .   .   .
     A   2   VAL   HG21   0.814     .   .   .
     A   2   VAL   HG22   0.814     .   .   .
     A   3   GLN   N      125.058   .   .   .

   stop_

save_
"""


# # noinspection PyUnusedLocal
def test_ppm_out_short_no_sequence(typer_app, using_nmrview, monkeypatch, clear_cache):

    monkeypatch.setattr("sys.stdin.isatty", lambda: False)

    path = path_in_test_data(__file__, "ppm_short.out")
    result = run_and_report(
        typer_app, [*NMRVIEW_IMPORT_SHIFTS, path], expected_exit_code=1
    )

    assert "ERROR" in result.stdout
    assert "couldn't read from stdin" in result.stdout


# noinspection PyUnusedLocal
def test_ppm_out_short(typer_app, using_nmrview, monkeypatch, clear_cache):

    monkeypatch.setattr("sys.stdin.isatty", lambda: False)

    STREAM = open(path_in_test_data(__file__, "ppm_short_seq.nef", local=True)).read()

    path = path_in_test_data(__file__, "ppm_short.out")
    result = run_and_report(typer_app, [*NMRVIEW_IMPORT_SHIFTS, path], input=STREAM)

    mol_sys_result = isolate_frame(result.stdout, SHIFTS_NMRPIPE)

    assert_lines_match(EXPECTED_PPM_OUT_SHORT, mol_sys_result)



import typer

import nef_pipelines
from nef_pipelines import nef_app

app = typer.Typer()
export_app = typer.Typer()

if nef_app.app:

    nef_app.app.add_typer(app, name="mars", help="- export mars [shifts and sequences]")

    app.add_typer(
        export_app, name="export", help="- export mars [shifts and sequences]"
    )

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.transcoders.mars.exporters.shifts  # noqa: F401



#!/usr/bin/env python3
#Importing libraries
import sys
import operator

import numpy as np
import pynmrstar
from pynmrstar import Entry
from dataclasses import dataclass
from random import sample
from itertools import tee
import argparse
parser = None



EXIT_ERROR = 1
def exit_error(msg):

        print(f'ERROR: {msg}')
        print('exiting...')
        sys.exit(EXIT_ERROR)


# https://stackoverflow.com/questions/312443/how-do-you-split-a-list-into-evenly-sized-chunks
def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

def select_frames(nef_target_data, target_types):

    for frame_name, frame_data in nef_target_data.frame_dict.items():

        for target_type in target_types:
            if frame_name.startswith(target_type):
                yield frame_name, frame_data


def add_random_errors_to_shifts(nef_target_data, atom_type_shift_errors):

    target_types = ['nef_chemical_shift_list']

    for _, frame_data in select_frames(nef_target_data, target_types):

        for loop_name, loop_data in frame_data.loop_dict.items():
            loop_tags = {tag:i for i,tag in enumerate(loop_data.tags)}

            sequence_code_index = loop_tags['sequence_code']
            chain_code_index = loop_tags['chain_code']
            atom_name_index = loop_tags['atom_name']
            shift_index = loop_tags['value']

            print('#')
            print('# fuzzer: adding shift errors ')
            print('#')

            for line_number, line in enumerate(loop_data):
                atom_name = line[atom_name_index]

                if atom_name in atom_type_shift_errors:
                    if atom_type_shift_errors[atom_name] > np.finfo(float).eps:

                        shift = float(line[shift_index])

                        atom_name = line[atom_name_index]

                        name = f'{line[chain_code_index]}.{line[sequence_code_index]}.{line[atom_name_index]}'
                        error = np.random.normal(scale = atom_type_shift_errors[atom_name])[0]

                        new_shift = shift + error

                        print(f'# add error  {error:4.3} to {name} {shift:10.7} -> {new_shift:10.7}')

                        line[shift_index] = "%10.7f" % new_shift




                            
            


# # remove random shifts



def add_random_remove_shifts(nef_target_data, atom_remove_percentage):
    
    target_frames = ['nef_chemical_shift_list']
    for frame_name, frame_data in select_frames(nef_target_data):

        target_type = None
        for target_frame in target_frames:
            if frame_name.startswith(target_frame):
                target_type = target_frame
                break


        if target_type == 'nef_chemical_shift_list':

            for loop_name, loop_data in frame_data.loop_dict.items():
                loop_tags = {tag:i for i,tag in enumerate(loop_data.tags)}

                sequence_code_index = loop_tags['sequence_code']
                chain_code_index = loop_tags['chain_code']
                residue_name_index = loop_tags['residue_name']
                atom_name_index = loop_tags['atom_name']
                shift_index = loop_tags['value']

                possible_lines_for_removal = {}

                for line_number, line in enumerate(loop_data):
                    atom_name = line[atom_name_index]

                    for atom_remove_name in atom_remove_percentage.keys():
                        if atom_remove_name in atom_name:
                            possible_lines_for_removal.setdefault(atom_remove_name,[]).append(line)


                lines_for_removal = []

                for atom_name, values in possible_lines_for_removal.items():
                    pass
                    # pass random.sample(atom_name)
                    # remove = lines_for_removal ??

                for line_for_removal in lines_for_removal:
                    loop_data.remove(line_for_removal)





# # 

def swap_shifts(nef_target_data, atom_swap_shifts_percentage):


    target_frames = ['nef_chemical_shift_list']
    for frame_name, frame_data in nef_target_data.frame_dict.items():


        target_type = None
        for target_frame in target_frames:
            if frame_name.startswith(target_frame):
                target_type = target_frame
                break


        if target_type == 'nef_chemical_shift_list':

            for loop_name, loop_data in frame_data.loop_dict.items():
                loop_tags = {tag:i for i,tag in enumerate(loop_data.tags)}

                sequence_code_index = loop_tags['sequence_code']
                chain_code_index = loop_tags['chain_code']
                residue_name_index = loop_tags['residue_name']
                atom_name_index = loop_tags['atom_name']
                value_index = loop_tags['value']

                possible_lines_for_swapping_shifts = {}

                for line_number, line in enumerate(loop_data):
                    atom_name = line[atom_name_index]

                    for atom_swap_name in atom_swap_shifts_percentage.keys():
                        if atom_swap_name in atom_name:
                            possible_lines_for_swapping_shifts.setdefault(atom_swap_name,[]).append(line)


                lines_for_swapping_shifts = {}

                for atom_name, values in possible_lines_for_swapping_shifts.items():
                    number_to_swap = int(atom_swap_shifts_percentage[atom_name] * len(values))
                    print('#')
                    print(f'# fuzzer number to swap {atom_name}  = {number_to_swap}')
                    print('#')
                    to_swap = sample(values, number_to_swap)

                    for pair in chunks(to_swap, 2):
                        if len(pair) > 1:
                            a, b = pair

                            name_a = f'{a[chain_code_index]} {a[sequence_code_index]} {a[atom_name_index]} {a[residue_name_index]}'
                            name_b = f'{b[chain_code_index]} {b[sequence_code_index]} {b[atom_name_index]} {b[residue_name_index]}'
                            print(f'# fuzzer swapping shift values for {name_a} to {name_b}')
                            a[value_index], b[value_index] = b[value_index], a[value_index]

                    print('#')

def read_args():

    parser = argparse.ArgumentParser(description='Add chemical shift errors to a NEF file')
    parser.add_argument('-a', '--alpha', metavar='ALPHA', type=float, nargs=1, default=0.0, dest='alpha',
                        help='an error in alpha chemical shifts')
    parser.add_argument('-b', '--beta', metavar='BETA', type=float, nargs=1, default=0.0, dest='beta',
                        help='an error in alpha chemical shifts')
    parser.add_argument(metavar='FILES', nargs=argparse.REMAINDER, dest='files')

    return parser.parse_args()


if __name__ == '__main__':

    args = read_args()

    is_tty = sys.stdin.isatty()

    if is_tty and len(args.files) == 0:
        parser.print_help()
        print()
        msg = """I require at least 1 argument or input stream with a chemical_shift_list frame"""

        exit_error(msg)

    entries = []
    try:
        if len(args.files) == 0:
            lines = sys.stdin.read()
            entries.append(Entry.from_string(lines))
        else:
            for file in args.files:
                entries.append(Entry.from_file(file))
    except OSError as e:
        msg = f"couldn't open target nef file because {e}"
        exit_error(msg)

    for entry in entries:
        add_random_errors_to_shifts(entry, {'CA': args.alpha, 'CB': args.beta})

        print(entry)

    #this is how often you should swap with another residue in % /100 you can define your own values here
    # atom_swap_shifts_percentage =  {'CA' : 0.10, 'CB': 0.05}
    # swap_shifts(nef_target_data, atom_swap_shifts_percentage)
    #
    #
    # # Add errors to shifts....
    # # these are the sd of the errors to add
    # atom_name_errors = {'CA': 0.0, 'CB': 0.1}
    # add_random_errors_to_shifts(nef_file_data, atom_name_errors)
    #
    # #this is how often you should remove a shift in % /100 you can define your own values here
    # atom_remove_percentage =  {'CA' : 0.05, CB: 0.10}
    # add_random_remove_shifts(nef_target_data, atom_remove_percentage)

    # print(nef_file_data)
#  




from argparse import Namespace
from pathlib import Path
from typing import Dict, List, Tuple

import typer
from pynmrstar import Entry, Saveframe

from nef_pipelines.lib.sequence_lib import chain_code_iter
from nef_pipelines.lib.shift_lib import shifts_to_nef_frame
from nef_pipelines.lib.structures import SequenceResidue
from nef_pipelines.lib.typer_utils import get_args
from nef_pipelines.lib.util import (
    cached_file_stream,
    exit_error,
    get_pipe_file_or_exit,
    process_stream_and_add_frames,
)
from nef_pipelines.transcoders.nmrview import import_app
from nef_pipelines.transcoders.nmrview.nmrview_lib import parse_shifts, read_sequence

app = typer.Typer()


# noinspection PyUnusedLocal
@import_app.command(no_args_is_help=True)
def shifts(
    chain_codes: str = typer.Option(
        "A",
        "--chains",
        help="chain codes as a list of names spearated by dots",
        metavar="<CHAIN-CODES>",
    ),
    entry_name: str = typer.Option("nmrview", help="a name for the entry"),
    pipe: Path = typer.Option(
        None,
        metavar="|PIPE|",
        help="pipe to read NEF data from, for testing [overrides stdin !use stdin instead!]",
    ),
    file_names: List[Path] = typer.Argument(
        ..., help="input files of type nmrview.out", metavar="<NMRVIEW-shifts>.out"
    ),
):
    """convert nmrview shift file <nmrview-shifts>.out to NEF"""

    try:
        args = get_args()

        process_shifts(args)
    except Exception as e:
        exit_error(f"reading sequence failed because {e}", e)


#
#
def process_shifts(args: Namespace):
    nmrview_frames = []

    for file_name, chain_code in zip(
        args.file_names, chain_code_iter(args.chain_codes)
    ):

        with cached_file_stream(file_name) as lines:

            sequence = _get_sequence_or_exit(args)

            chain_seqid_to_type = _sequence_to_residue_type_lookup(sequence)

            nmrview_shifts = parse_shifts(
                lines, chain_seqid_to_type, chain_code=chain_code, file_name=file_name
            )

            frame = shifts_to_nef_frame(nmrview_shifts, args.entry_name)

            nmrview_frames.append(frame)

    entry = process_stream_and_add_frames(nmrview_frames, args)

    print(entry)


def sequence_from_frames(frames: Saveframe) -> List[SequenceResidue]:

    residues = []
    for frame in frames:
        for loop in frame:
            chain_code_index = loop.tag_index("chain_code")
            sequence_code_index = loop.tag_index("sequence_code")
            residue_name_index = loop.tag_index("residue_name")

            for line in loop:
                chain_code = line[chain_code_index]
                sequence_code = int(line[sequence_code_index])
                residue_name = line[residue_name_index]
                residue = SequenceResidue(chain_code, sequence_code, residue_name)
                residues.append(residue)

    return residues


# TODO this should be replaced by a library function from nef or sequence utils...
# also need ro report what file we are trying to read shifts from
def _get_sequence_or_exit(args):
    sequence_file = None
    if "sequence" in args:
        sequence_file = args.sequence

    sequence = None
    if not sequence_file:
        try:
            stream = get_pipe_file_or_exit(args)

            entry = Entry.from_file(stream)
            frames = entry.get_saveframes_by_category("nef_molecular_system")
            sequence = sequence_from_frames(frames)

        except Exception as e:
            exit_error(f"failed to read sequence from input stream because {e}", e)

    else:
        with open(sequence_file, "r") as lines:
            sequence = read_sequence(lines, chain_code=args.chain_code)
    return sequence


def _sequence_to_residue_type_lookup(
    sequence: List[SequenceResidue],
) -> Dict[Tuple[str, int], str]:
    result: Dict[Tuple[str, int], str] = {}
    for residue in sequence:
        result[residue.chain_code, residue.sequence_code] = residue.residue_name
    return result


if __name__ == "__main__":

    typer.run(shifts)
